// https://liquidsdr.org/blog/modem-ber-sim : simulate modem bit error rate
#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <math.h>
#include <liquid/liquid.h>

int main() {
    // simulation parameters
    modulation_scheme ms    = LIQUID_MODEM_QPSK; // modulation scheme
    float SNRdB_min         = -5.0f;             // starting SNR value
    float SNRdB_max         = 30.0f;             // ending SNR value
    float SNRdB_step        =  1.0f;             // step size
    unsigned int num_trials = 1e6;               // number of symbols

    // create modem objects
    modem mod = modem_create(ms);   // modulator
    modem dem = modem_create(ms);   // demodulator
    
    // compute derived values and initialize counters
    unsigned int  m = modem_get_bps(mod);       // modulation bits/symbol
    unsigned int  M = 1 << m;                   // constellation size: 2^m
    unsigned int  i, sym_in, sym_out;
    float complex sample;

    // iterate through SNR values
    printf("# modulation scheme : %s\n", modulation_types[ms].name);
    printf("# %8s %8s %8s %12s\n", "SNR [dB]", "errors", "trials", "BER");
    float SNRdB = SNRdB_min;    // set SNR value
    while (1) {
        // compute noise standard deviation
        float nstd = powf(10.0f, -SNRdB/20.0f);

        // reset modem objects (only necessary for differential schemes)
        modem_reset(mod);
        modem_reset(dem);

        // run trials
        unsigned int num_bit_errors = 0;
        for (i=0; i<num_trials; i++) {
            // generate random input symbol and modulate
            sym_in = rand() % M;
            modem_modulate(mod, sym_in, &sample);

            // add complex noise and demodulate
            sample += nstd*(randnf() + _Complex_I*randnf())/sqrtf(2.0f);
            modem_demodulate(dem, sample, &sym_out);

            // count errors
            num_bit_errors += count_bit_errors(sym_in, sym_out);
        }

        // compute results and print formatted results to screen
        unsigned int num_bit_trials = m * num_trials;
        float ber = (float)num_bit_errors / (float)(num_bit_trials);
        printf("  %8.2f %8u %8u %12.4e\n", SNRdB, num_bit_errors, num_bit_trials, ber);

        // stop iterating if SNR exceed maximum or no errors were detected
        SNRdB += SNRdB_step;
        if (SNRdB > SNRdB_max || num_bit_errors == 0)
            break;
    }

    // destroy modem objects and return
    modem_destroy(mod);
    modem_destroy(dem);
    return 0;
}
